package com.digistamp.tsp.demo;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Properties;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.bouncycastle.asn1.ASN1Encoding;
import org.bouncycastle.asn1.nist.NISTObjectIdentifiers;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.tsp.TimeStampToken;
import org.junit.Test;

import com.digistamp.tsp.ConfigManager;
import com.digistamp.tsp.TSAservice;
import com.digistamp.tsp.TsRequest;
import com.digistamp.tsp.TsResponse;
import com.digistamp.util.Constants;

/**
 * Time stamping the contents of a file and save the new time stamp to disk.
 * Internet communication to the DigiStamp server is performed as part of the
 * processing.
 * <p>
 * Edit the file <code>com.digistamp.properties</code> to input your DigiStamp
 * Account number and password before running this demo program. You may have an
 * account in DigiStamp test environment and one in Production. The contents of
 * <code>com.digistamp.properties</code> determines if the Test or Production
 * TSA nodes are called to create the timestamp.
 * <p>
 * This source code is include with the SecureTime Java API distribution.
 * <p>
 * Given a file name, the process performs:
 * <OL>
 * <LI>calculate SHA-512 hash of the contents
 * <LI>Construct a time stamp request
 * <LI>Internet communication to the time stamp server
 * <LI>Verify the time stamp signature
 * <LI>Verify the certificate trust
 * <LI>Writes the time stamp to disk
 * </OL>
 * 
 * Optional second parameter is the output file name. If not specified then the
 * output file name is the input name appended with .p7s. If the output file
 * exists then a random filename will be created based on the intended name.
 * 
 * @author Copyright  2015-2017 DigiStamp
 * @version 4.3.3
 */
public class CreateTimestamp {

	private static Logger logger = Logger.getLogger(CreateTimestamp.class.getName());
	/**
	 * you can use as singleton because this is thread safe.
	 */
	public TSAservice theTSAservice = new TSAservice();

	/**
	 * Execute the demo from a command line. Enter:
	 * <OL>
	 * <LI>input file name that will be time stamped
	 * </OL>
	 * 
	 * @see CreateTimestamp#getTimeStamp(InputStream,String) for details
	 */
	public static void main(String args[]) {

		try {
			int argc = args.length;
			if (argc < 1) {
				System.out.println("Please enter the name of the file to be timestamped.");
				System.out.println("  The filename needs to be the full path or if just the name");
				System.out.println("  will assume the current directory.");
				System.out.println("The demo will calculate the SHA-512 hash of your file and then");
				System.out.println("  request a timestamp for that hash value.");
				System.exit(1);
			}

			/*
			 * Problems? Uncomment below to see logging details
			 */
//			 enableLogging();

			logger.fine("DigiStamp Java Toolkit version: ");
			logger.fine(TSAservice.VERSION);
			logger.fine(" Java Version: " + System.getProperty("java.version") + " from "
					+ System.getProperty("java.vendor"));
			logger.fine("user.dir: " + System.getProperty("user.dir"));

			// add the BouncyCastle JCE
			int bouncyCastleIsLoaded = java.security.Security.addProvider(new BouncyCastleProvider());
			if (bouncyCastleIsLoaded == -1)
				logger.warning(
						"BouncyCastleProvider addProvider maybe problem, got: " + String.valueOf(bouncyCastleIsLoaded));
			else
				logger.fine("BouncyCastleProvider addProvider at: " + String.valueOf(bouncyCastleIsLoaded));

			String filename = args[0];
			FileInputStream fis = null;
			File theInputFile = new File(filename);
			try {
				fis = new FileInputStream(theInputFile);
			} catch (IOException e) {
				System.out.println("Failure because unable to read file to be protected, " + filename);
				System.out.println(e);
				System.exit(1);
			}

			FileOutputStream out = null;
			String outFileName = null;
			if (argc == 2) {
				outFileName = args[1];
			} else {
				outFileName = filename + ".p7s";
			}

			File theOutputFile = new File(outFileName);
			if (theOutputFile.exists()) {
				theOutputFile = new File(theOutputFile.getAbsolutePath());
				theOutputFile = File.createTempFile(filename, ".p7s", theOutputFile.getParentFile());
			}

			CreateTimestamp theTSAservice = new CreateTimestamp();
			TimeStampToken aTimeStampJavaObject = theTSAservice.getTimeStamp(fis, theInputFile.getName());
			fis.close();
			if (aTimeStampJavaObject == null) {
				System.exit(1);
			}
			byte[] encoded3161timestamp = aTimeStampJavaObject.toCMSSignedData().toASN1Structure()
					.getEncoded(ASN1Encoding.DL);

			try {
				out = new FileOutputStream(theOutputFile);
			} catch (FileNotFoundException e) {
				System.out.println("Failure because unable to save returned timestamp");
				System.out.println(e);
				System.exit(1);
			}

			try {
				out.write(encoded3161timestamp);
				out.close();
				System.out.println("Succesful. Timestamp saved as " + theOutputFile.getName());
				System.exit(0);

			} catch (IOException e) {
				System.out.println("Failure because unable to write returned timestamp");
				System.out.println(e);
				System.exit(1);
			}
		} catch (Exception exception) {
			System.out.println("Unexpected exception in processing");
			System.out.println(exception);
			exception.printStackTrace();
			System.out.println("details: " + exception.getMessage());
			System.exit(1);
		}
	}

	static void enableLogging() {
		Handler[] handlers = java.util.logging.Logger.getLogger("").getHandlers();
		for (int index = 0; index < handlers.length; index++) {
			handlers[index].setLevel(Level.FINEST);
		}
		java.util.logging.Logger theLogger = Logger.getLogger("com.digistamp");
		theLogger.setLevel(Level.FINEST);
	}

	/**
	 * A demonstration of time stamp the content of a file.
	 * 
	 * The configuration of Internet communications will is done automatically /
	 * independently because this toolkit will initialize by reading the
	 * <code>com.digistamp.properties</code>. It is in that file you will have
	 * put your DigiStamp user ID and password.
	 */
	public TimeStampToken getTimeStamp(InputStream toBeTimeStamped, String filename) throws IOException {

		// we are going to calculate that hash value and then pass that hash to
		// the toolkit.
		MessageDigest md = null;
		try {
			md = MessageDigest.getInstance("SHA-512");
		} catch (NoSuchAlgorithmException e) {
			System.out.println("Failure because unable to intialize SHA-512 digester");
			System.out.println(e);
			return null;
		}
		byte[] dataBytes = new byte[8192];
		int nread = 0;
		while ((nread = toBeTimeStamped.read(dataBytes)) != -1) {
			md.update(dataBytes, 0, nread);
		}
		byte[] mdbytes = md.digest();

		logger.fine("This is the current configuration: ");
		logger.fine(theTSAservice.toString());
		// create the request with the pre-calculated hash and tell the toolkit
		// which hash algorithm you used
		TsRequest request = new TsRequest(mdbytes, NISTObjectIdentifiers.id_sha512);
		request.setFilename(filename);
		TsResponse response = theTSAservice.getTimestamp(request);
		if (response.isGoodResponse()) {
			System.out.println("Timestamp retrieved from common name: " + response.getOriginOfTimestamp());
			return response.getTimeStampResponse().getTimeStampToken();
		} else {
			System.out.println("Failure because " + response.getStatusString());
			return null;
		}

	}

	public String alternateUserID = "aUserID";
	public String alternatePassword = "thePassword";

	@Test
	public void testConfigurationAlternatives() {
		// you can use as singleton because this is thread safe
		TSAservice theTSAservice;
		try {

			TsResponse response = null;

			// an arbitrary SHA512 value is 64 bytes
			TsRequest request = new TsRequest(new byte[64], NISTObjectIdentifiers.id_sha512.getId());

			theTSAservice = new TSAservice();
			assertTrue("expected to find the default value of 5000",
					theTSAservice.getConfiguration().getProperty(Constants.HTTP_CONNECTION_TIMEOUT, "").equals("5000"));

			/*
			 * To run this test create your com.digistamp.properties file and
			 * put in only your DigiStamp user ID and password. Also, you may
			 * want to use the TEST environment, do that by adding this to the
			 * properties file: com.digistamp.pki.Test.Untrusted=true
			 * 
			 * This demonstrates where the configuration comes primarily from
			 * API toolkit defaults. Then, overlaid with configuration items in
			 * file com.digistamp.properties. Each timestamp transaction will
			 * share this configuration.
			 * 
			 * RECOMMENDED method.
			 */
			String readUser = theTSAservice.getConfiguration().getProperty(Constants.HTTP_USER, "");
			String readPassword = theTSAservice.getConfiguration().getProperty(Constants.HTTP_PASSWORD, "");
			assertFalse("did not find user ID in com.digistamp.properties",
					readUser.equals("") || readPassword.equals(""));
			response = theTSAservice.getTimestamp(request);
			assertTrue("did not return a valid timestamp because: " + response.getStatusString(),
					response.isGoodResponse());

			/*
			 * Use java code to set the configuration of the TSA service. Each
			 * timestamp transaction will share this new default configuration.
			 * 
			 * RECOMMENDED method if you deploy this toolkit using JAR files
			 * exclusively. This takes the place of using
			 * <code>com.digistamp.properties</code> XML file to set your
			 * configuration.
			 * 
			 */

			theTSAservice = new TSAservice(); // start test with new instance
			assertTrue("default was set to false",
					theTSAservice.getConfiguration().getProperty(Constants.HTTP_ROUND_ROBIN, "").equals("false"));
			Properties mySpecialConfigs = new Properties();
			mySpecialConfigs.putAll(ConfigManager.getDefaultProperties());
			mySpecialConfigs.put(Constants.HTTP_ROUND_ROBIN, "true");
			// set to load the CA ROOT certificates by using predefined names in
			// classpath jar files. Retrieve the certificates in a JAR file
			// https://www.digistamp.com/pubkeys/com.digistamp.root.jar
			// or, when using the test environment:
			// https://www.digistamp.com/pubkeys/com.digistamp.root.TEST.jar
			mySpecialConfigs.put(Constants.ROOT_CERTS_BY_NAME, "true");

			// are you using the test environment? if yes, testing, then add:
			mySpecialConfigs.put(Constants.PKI_UNTRUSTED, "true");

			// set your credentials, unless you setup IP Based authentication
			// mySpecialConfigs.put(Constants.HTTP_USER,"yourUserAccountOrEmailAddress");
			// mySpecialConfigs.put(Constants.HTTP_PASSWORD,"passwordYouChose");

			theTSAservice.setConfiguration(mySpecialConfigs);

			response = theTSAservice.getTimestamp(request);
			assertTrue("did not return a valid timestamp because: " + response.getStatusString(),
					response.isGoodResponse());
			assertTrue("changed the value",
					theTSAservice.getConfiguration().getProperty(Constants.HTTP_ROUND_ROBIN, "").equals("true"));

			/*
			 * Use the current TSA Service configuration - you have two methods
			 * above to do that. Now, override just which user account is
			 * associated with this timestamp request. You will need to create a
			 * "sub-account" on your Browser account interface to run this test.
			 * 
			 * RECOMMENDED method if case of having multiple accounts at the TSA
			 * service provider
			 */
			theTSAservice = new TSAservice(); // start test with new object
			String preExistingConfigUser = theTSAservice.getConfiguration().getProperty(Constants.HTTP_USER, "");
			assertFalse("configuration problem, these need to be different for this test",
					preExistingConfigUser.equals(alternateUserID));
			response = theTSAservice.getTimestamp(request, alternateUserID, alternatePassword);
			assertTrue("response was for user not as expected", response.getUser().equals(alternateUserID));
			/*
			 * the defaults we set previously should still be in place
			 */
			response = theTSAservice.getTimestamp(request);
			assertTrue("did not return a valid timestamp because: " + response.getStatusString(),
					response.isGoodResponse());
			assertTrue("response was for user not as expected", response.getUser().equals(preExistingConfigUser));

			// want to see the progress of request processing
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

}
