package com.digistamp.tsp.demo;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.TimeZone;
import java.util.logging.Logger;

import org.bouncycastle.cms.CMSException;
import org.bouncycastle.cms.CMSSignedData;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.operator.DigestCalculator;
import org.bouncycastle.operator.DigestCalculatorProvider;
import org.bouncycastle.operator.OperatorCreationException;
import org.bouncycastle.operator.bc.BcDigestCalculatorProvider;
import org.bouncycastle.tsp.TimeStampToken;
import org.bouncycastle.tsp.TimeStampTokenInfo;
import org.bouncycastle.tsp.cms.CMSTimeStampedData;
import org.bouncycastle.tsp.cms.CMSTimeStampedDataGenerator;
import org.bouncycastle.tsp.cms.ImprintDigestInvalidException;
import org.bouncycastle.util.encoders.Hex;

import com.digistamp.tsp.TSAservice;

/**
 * Verifying a timestamp and either displaying its hash value, or confirming it
 * against the protected file, and displaying the protected time. No Internet
 * communications occur during this process.
 * <p>
 * This source code is include with the SecureTime Java API distribution.
 * <p>
 * Given a file name, the process performs:
 * <OL>
 * <LI>read the timestamp provided
 * <LI>verify the provided timestamp signature
 * <LI>display the protected hash value in HEX encoding
 * <LI>verify any provided protected file against this hash value
 * <LI>provide the protected time in the timestamp with relevant statement about
 * the authenticity of any protected file
 * </OL>
 * 
 * TODO: this logic consumes too much memory for hash calculation.
 * 
 * @author Copyright  2015 DigiStamp
 * @version 4.1
 */
public class VerifyTimestamp {

	private static Logger logger = Logger.getLogger(VerifyTimestamp.class.getName());

	private static SimpleDateFormat dateFormatGmt = new SimpleDateFormat("yyyy-MMM-dd HH:mm:ss zzz");

	static {
		dateFormatGmt.setTimeZone(TimeZone.getTimeZone("GMT"));
	}

	/**
	 * you can use as singleton because this is thread safe.
	 */
	public TSAservice theTSAservice = new TSAservice();

	/**
	 * Execute the demo from a command line. Enter:
	 * <OL>
	 * <LI>input - file name of the timestamp, usually a .p7s
	 * <LI>(optionally) input - the file that was timestamped
	 * </OL>
	 * 
	 */
	public static void main(String args[]) {
		try {
			int argc = args.length;
			if (argc < 1) {
				System.out.println(
						"Please enter the name of timestamp to be verified, and optionally the file it protects.");
				System.out.println("  The timestamp name needs to be the full path or, if just the name,");
				System.out.println("  will assume the current directory.");
				System.out.println("The demo will verify your timestamp, and provide the timestamp 'time'");
				System.out.println("  of the protected file you indicated.");
				System.exit(1);
			}

			logger.fine("DigiStamp Java Toolkit version: ");
			logger.fine(TSAservice.VERSION);
			logger.fine(" Java Version: " + System.getProperty("java.version") + " from "
					+ System.getProperty("java.vendor"));
			logger.fine("user.dir: " + System.getProperty("user.dir"));

			int bouncyCastleIsLoaded = java.security.Security.addProvider(new BouncyCastleProvider());
			if (bouncyCastleIsLoaded == -1)
				logger.warning(
						"BouncyCastleProvider addProvider maybe problem, got: " + String.valueOf(bouncyCastleIsLoaded));
			else
				logger.fine("BouncyCastleProvider addProvider at: " + String.valueOf(bouncyCastleIsLoaded));

			String timestampname = null, protectedfilename = null;

			if (argc == 2) {
				// as convenience for user that reverses the parameter (e.g. the
				// windows "sendTo" function)
				String parm2ending = args[1].substring(args[1].length() - 4, args[1].length());
				if (parm2ending.equalsIgnoreCase(".p7s")) {
					timestampname = args[1];
					protectedfilename = args[0];
				} else {
					timestampname = args[0];
					protectedfilename = args[1];
				}
			} else {
				timestampname = args[0];
			}

			// Read the timestamp
			FileInputStream theTimestampAsRFC3161 = null;
			File theTimestampFile = new File(timestampname);
			try {
				theTimestampAsRFC3161 = new FileInputStream(theTimestampFile);
			} catch (IOException e) {
				System.out.println("Failure because unable to read the timestamp file: " + timestampname);
				System.out.println(e);
				System.exit(1);
			}
			CMSSignedData aGenericCMSSignedData = new CMSSignedData(theTimestampAsRFC3161);
			TimeStampToken aTimestamp = new TimeStampToken(aGenericCMSSignedData);

			FileInputStream theProtectedData = null;
			File theProtectedFile = null;
			if (protectedfilename != null) {
				theProtectedFile = new File(protectedfilename);
				try {
					theProtectedData = new FileInputStream(theProtectedFile);
				} catch (IOException e) {
					System.out.println("Failure because unable to read the protected file: " + protectedfilename);
					System.out.println(e);
					System.exit(1);
				}
			}

			VerifyTimestamp verifyUtility = new VerifyTimestamp();
			verifyUtility.validate(aTimestamp, theProtectedData);
			TimeStampTokenInfo theSignedPayload = aTimestamp.getTimeStampInfo();
			String HashAlgorithm = theSignedPayload.getHashAlgorithm().getAlgorithm().toString();
			String HashValue = new String(Hex.encode(theSignedPayload.getMessageImprintDigest()));
			String protectionTime = theSignedPayload.getGenTime().toString();
			System.out.println(protectionTime);
			System.out.print(dateFormatGmt.format(theSignedPayload.getGenTime()));
			System.out.println(" Protected hash (" + HashAlgorithm + "): " + HashValue);
			System.exit(0);

		} catch (Exception exception) {
			System.out.println("Unexpected exception in processing");
			System.out.println(exception);
			exception.printStackTrace();
			System.out.println("details: " + exception.getMessage());
			System.exit(1);
		}
	}

	/**
	 * Does the hash of the content data match what is recorded in the
	 * timestamp? Calculate the digest / hash value of the
	 * "protect data (content)". Use the digest method defined in the timestamp.
	 * 
	 * from: org.bouncycastle.tsp.testCMSTimeStampedDataTest
	 */
	public static void testValidateAllTokens(CMSTimeStampedData cmsTimeStampedData)
			throws OperatorCreationException, IOException, CMSException, ImprintDigestInvalidException {
		DigestCalculatorProvider digestCalculatorProvider = new BcDigestCalculatorProvider();

		DigestCalculator imprintCalculator = cmsTimeStampedData
				.getMessageImprintDigestCalculator(digestCalculatorProvider);

		imprintCalculator.getOutputStream().write(cmsTimeStampedData.getContent());

		byte[] digest = imprintCalculator.getDigest();

		TimeStampToken[] tokens = cmsTimeStampedData.getTimeStampTokens();
		for (int i = 0; i < tokens.length; i++) {
			cmsTimeStampedData.validate(digestCalculatorProvider, digest, tokens[i]);
		}
	}

	/**
	 * Do all validations
	 */
	public void validate(TimeStampToken aTimestamp, FileInputStream theProtectedData)
			throws CMSException, OperatorCreationException, IOException, ImprintDigestInvalidException {

		validateDigestValue(aTimestamp, theProtectedData);

		// TODO now need to validate the signature ( CRLs etc)

	}

	/**
	 * Does the provided timestamp contain the hash of the provided
	 * "protected data"? Throw an exception if any issues.
	 * 
	 */
	public void validateDigestValue(TimeStampToken aTimestamp, FileInputStream theProtectedData)
			throws CMSException, OperatorCreationException, IOException, ImprintDigestInvalidException {
		CMSTimeStampedDataGenerator aUtility = new CMSTimeStampedDataGenerator();
		CMSTimeStampedData aCMSTimeStampedData = null;
		if (theProtectedData != null) {
			aCMSTimeStampedData = aUtility.generate(aTimestamp, theProtectedData);
			testValidateAllTokens(aCMSTimeStampedData);
		}
	}

}
